# coding: utf-8

from django.core.management.base import BaseCommand, CommandError

from ...index import get_indexer


class Command(BaseCommand):
    help = 'Updates elasticsearch index'

    def handle(self, *args, **options):
        indexer = get_indexer()
        indexer.reset_index()
        indexer.index_documents()
        indexer.refresh_index()



# coding: utf-8

"""
Configuration for django-el.


CONNECTIONS - a dictionary which keys are conneciton aliases, and it's values
are elasticsearc_dsl.connections.create_connection arguments.
"""

from django.conf import settings
from django.utils.module_loading import import_string

from elasticsearch_dsl.connections import connections


DEFAULT_CONNECTIONS = {
    'default': {
        'hosts': ['localhost:9200'],
    }
}
INDEX_NAME = getattr(settings, 'ELASTICSEARCH_INDEX_NAME', 'elastic')
CONNECTIONS = getattr(settings, 'ELASTICSEARCH_CONNECTIONS', DEFAULT_CONNECTIONS)


def create_connections():
    """Create connections to elasticsearch as defined in settings.py."""
    for alias, params in CONNECTIONS.items():
        processed_params = {}
        for param_name, param_value in params.items():

            if param_name == 'serializer' and isinstance(param_value, str):
                serializer_class = import_string(param_value)
                param_value = serializer_class()

            processed_params[param_name] = param_value

        connections.create_connection(alias, **processed_params)



# coding: utf-8

from django.db.models import signals

from . import index


def get_indexed_instance(instance):
    indexed_instance = instance.get_indexed_instance()
    if indexed_instance is None:
        return

    # Make sure that the instance is in its class's indexed objects
    if not type(indexed_instance).get_indexable().filter(
            pk=indexed_instance.pk).exists():
        return

    return indexed_instance


def post_save_signal_handler(instance, **kwargs):
    indexed_instance = get_indexed_instance(instance)

    if indexed_instance:
        index.get_indexer().add_document(indexed_instance)


def post_delete_signal_handler(instance, **kwargs):
    indexed_instance = get_indexed_instance(instance)

    if indexed_instance:
        index.get_indexer().delete_document(indexed_instance)


def register_signal_handlers():
    for model in index.get_indexed_models():
        signals.post_save.connect(post_save_signal_handler, sender=model)
        signals.post_delete.connect(post_delete_signal_handler, sender=model)



# coding: utf-8

from django.apps import AppConfig

from .signals import register_signal_handlers
from .conf import create_connections


class DefaultConfig(AppConfig):
    name = 'el'
    verbose_name = 'el'

    def ready(self):
        # импортировать сигналы для их регистрации
        import el.signals
        create_connections()
        register_signal_handlers()



# coding: utf-8

from django.apps import apps
from django.db.models.fields import FieldDoesNotExist

from elasticsearch_dsl import connections
from elasticsearch.helpers import streaming_bulk
from elasticsearch import NotFoundError

from .models import Indexed
from . import conf


def get_indexed_models():
    return [
        model for model in apps.get_models()
        if issubclass(model, Indexed) and not model._meta.abstract
    ]


class Indexer:
    def __init__(self, index):
        self.es = connections.connections.get_connection()
        self.index = index

    def reset_index(self):
        # Delete old index
        try:
            self.es.indices.delete(self.index)
        except NotFoundError:
            pass

        # Create new index
        self.es.indices.create(self.index)

    def refresh_index(self):
        self.es.indices.refresh(self.index)

    def save_mapping(self, model_class):
        mapping = model_class.get_mapping()
        mapping.save(self.index)
        print('Saved mapping {}\n'.format(mapping.doc_type))

    def index_documents(self):
        models = list(get_indexed_models())

        for model in models:
            self.save_mapping(model)

            model_instances = model.get_indexable().iterator()
            docs = (self.to_indexable_dict(d) for d in model_instances)
            for ok, info in streaming_bulk(self.es, docs):
                print("  Document with id %s indexed." % info['index']['_id'])

    def add_document(self, doc):
        self.es.index(
            self.index,
            doc.get_mapping().doc_type,
            self.to_indexable_dict(doc),
            doc.id,
        )

    def delete_document(self, doc):
        try:
            self.es.delete(
                self.index,
                doc.get_mapping().doc_type,
                doc.id,
            )
        except NotFoundError:
            pass

    def to_indexable_dict(self, obj):
        cls = obj.__class__
        fields = list(cls.get_mapping().properties.properties.to_dict().keys())

        data = {
            '_index': conf.INDEX_NAME,
            '_type': cls.get_mapping_name(),
            '_id': obj.pk,
        }
        for field_name in fields:
            if field_name == 'content_type':
                data['content_type'] = cls.indexed_get_content_type()
            elif field_name == 'pk':
                data['pk'] = obj.pk
            else:
                data[field_name] = self.get_value_from_field(field_name, obj)

        return data

    def get_value_from_field(self, field_name, obj):
        try:
            field = obj.__class__._meta.get_field(field_name)
            value = field._get_val_from_obj(obj)
        except FieldDoesNotExist:
            value = getattr(obj, field_name, None)
            if callable(value):
                value = value()

        return value


def get_indexer():
    return Indexer(conf.INDEX_NAME)



# coding: utf-8

from django.db import models

from elasticsearch_dsl import Mapping, String
from elasticsearch_dsl import Search


class DjangoObjectsSearch(Search):

    def __init__(self, *args, **kwargs):
        self.model_class = kwargs.pop('model_class', None)
        super().__init__(*args, **kwargs)

    def _clone(self, *args, **kwargs):
        s = super()._clone(*args, **kwargs)
        s.model_class = self.model_class
        return s

    def execute(self, cast=True, *args, **kwargs):
        hits = super().execute(*args, **kwargs)

        if not cast:
            return hits

        # Get pks from results
        pks = [str(hit.pk) for hit in hits]

        # Initialise results dictionary
        results = dict((pk, None) for pk in pks)

        # Find objects in database and add them to dict
        queryset = self.model_class.objects.filter(pk__in=pks)
        for obj in queryset:
            results[str(obj.pk)] = obj

        # Return results in order given by ElasticSearch
        return [results[str(pk)] for pk in pks if results[str(pk)]]


class Indexed(object):

    @classmethod
    def get_mapping_name(cls):
        return cls.__name__.lower()

    @classmethod
    def get_mapping(cls):
        m = Mapping(cls.get_mapping_name())
        m.field('pk', 'integer')
        m.field('content_type', String(index='not_analyzed'))
        return cls.configure_mapping(m)

    @classmethod
    def configure_mapping(cls, mapping):
        return mapping

    @classmethod
    def indexed_get_parent(cls, require_model=True):
        for base in cls.__bases__:
            if issubclass(base, Indexed) and (issubclass(base, models.Model) or require_model is False):
                return base

    @classmethod
    def indexed_get_content_type(cls):
        content_type = (cls._meta.app_label + '_' + cls.__name__).lower()

        # Get parent content type
        parent = cls.indexed_get_parent()
        if parent:
            parent_content_type = parent.indexed_get_content_type()
            return parent_content_type + '_' + content_type
        else:
            return content_type

    def get_indexed_instance(self):
        # This is accessed on save by the wagtailsearch signal handler, and in edge
        # cases (e.g. loading test fixtures), may be called before the specific instance's
        # entry has been created. In those cases, we aren't ready to be indexed yet, so
        # return None.
        try:
            return self.specific
        except self.specific_class.DoesNotExist:
            return None

    @classmethod
    def search(cls):
        ct = cls.indexed_get_content_type()
        return DjangoObjectsSearch(
            model_class=cls).filter('term', content_type=ct)

    @classmethod
    def get_indexable(cls):
        raise NotImplementedError
